#!/bin/python

import argparse
import os
from pathlib import Path
import random
import hashlib
import psutil
import numpy as np
import time

def time_rebuild(basedir, command):
    os.chdir(basedir)
    p = psutil.Popen(command, shell=True)
    cpuTimes = p.cpu_times()
    while True:
        time.sleep(1e-4)
        cpuTimes = p.cpu_times()
        retCode = p.poll()
        if retCode is not None:
            break

    return cpuTimes

# Random modification to the bottom of C/C++ and CUDA files
# that reliably forces recompilation
CPP_MOD = """
{inline} int dontOptimizeMeAway{hash}() {{
    volatile int a = 0;
    a += 1;
    if (a < 2) {{ while(true); }}
    return a;
}}
"""
def modify_file(f):
    # A weak but good-enough hash function
    file_hash = int(hashlib.sha256(f.encode('utf-8')).hexdigest(), 16) % 10**8
    ext = os.path.splitext(f)[1]
    if ext in ['.h', '.hpp', '.cuh', '.c', '.cpp', '.cu']:
        with open(f, 'a') as _f:
            _f.write(CPP_MOD.format(hash=file_hash, inline=('inline' if ext in ['.h', '.hpp', '.cuh'] else '')))

def sample_files(k, filedata):
    if k > len(filedata):
        raise Exception('sample_files - provided k is smaller than the number of available files')
    a = np.array(list(filedata.keys()))
    normalized = a / a.sum(0)
    return np.random.choice(list(filedata.values()), p=normalized, size=k, replace=False)
    

def get_file_length(path):
    return sum(1 for line in open(path))

def get_files_with_extensions(dirs, extensions):
    all_files = []
    for d in dirs:
        for root, dirs, files in os.walk(d):
            all_files.extend([os.path.join(root, f) for f in files])
    def is_ext(x):
        _, ext = os.path.splitext(x)
        return ext[1:] in extensions
    files = filter(lambda x: is_ext(x), all_files)
    return files

def main():
    parser = argparse.ArgumentParser(description='Randomly augment a source file in a project.')
    parser.add_argument('--basedir', type=str, required=True, help='The base directory in project')
    parser.add_argument('-d', '--dir', action='append', nargs=1, help='Directories eligible for random edit')
    parser.add_argument('-e', '--extension', action='append', nargs=1, help='List of extensions eligible for random edit')
    parser.add_argument('--builddir', type=str, required=True, help='The project directory in which to run the build command')
    parser.add_argument('--numsamples', type=int, required=True, help='The number of samples to take')
    parser.add_argument('--command', type=str, required=True, help='Command to rebuild things')
    parser.add_argument('--logfile', type=str, required=True, help='Logfile to write results things')
    

    args = parser.parse_args()
    dir_args = [d[0] for d in args.dir]
    extensions = [e[0] for e in args.extension]
    print(args)

    search_dirs = []

    for d in dir_args:
        concat = os.path.join(args.basedir, d)
        if not os.path.isdir(concat):
            raise Exception(f"Directory {d} doesn't exist in {args.basedir}")
        search_dirs.append(concat)
l
    files = get_files_with_extensions(dirs=search_dirs, extensions=extensions)
    file_to_length = dict(sorted({get_file_length(f): f for f in files}.items()))
    samples = sample_files(k=args.numsamples, filedata=file_to_length)

    with open(args.logfile, 'w') as out:
        for f in samples:
            print(f"Modifying file {f}")
            modify_file(f)
            cpu_time = time_rebuild(basedir=args.builddir, command=args.command)
            usr_cpu_time = cpu_time.children_user
            print(f"Rebuild of {f} completed in {usr_cpu_time}")
            out.write(f"{f} {usr_cpu_time}\n")
            
if __name__ == "__main__":
    main()
