/*
 * Copyright (c) Facebook, Inc. and its affiliates.
 *
 * This source code is licensed under the BSD-style license found in the
 * LICENSE file in the root directory of this source tree.
 */

#include <chrono>
#include <cmath>
#include <cstdlib>

#include <gflags/gflags.h>
#include <glog/logging.h>

#include "flashlight/app/asr/criterion/ConnectionistTemporalClassificationCriterion.h"
#include "flashlight/app/common/Runtime.h"
#include "flashlight/ext/common/DistributedUtils.h"
#include "flashlight/fl/contrib/contrib.h"

#include "asr/InputSizes.h"
#include "asr/TargetSizes.h"

DEFINE_bool(distributed_enable, true, "Enable distributed training");
DEFINE_int64(
    distributed_max_devices_per_node,
    8,
    "the maximum number of devices per training node");
DEFINE_int64(
    distributed_world_rank,
    0,
    "rank of the process (Used if distributed_rndv_filepath is not empty)");
DEFINE_int64(
    distributed_world_size,
    1,
    "total number of the process (Used if distributed_rndv_filepath is not empty)");
DEFINE_string(
    distributed_rndv_filepath,
    "",
    "Shared file path used for setting up rendezvous."
    "If empty, uses MPI to initialize.");
DEFINE_int64(batch_size, 10, "Batch size");
DEFINE_int64(num_transformer_layers, 36, "Number of transformer layers");
DEFINE_int64(num_flt_bank, 80, "# of filter banks (well, fake)");
DEFINE_int64(num_labels, 31, "# of labels (well, fake)");

using namespace fl;

// ]--- Benchmark parameters
const unsigned numWarmupIters = 300;
const unsigned numItersPerSample = 100;
const unsigned numSamplesToBench = 5;

// Model parameters
const int dModel = 768; // context size
const int nHead = 4; // number of attention heads
const int dimFeedforward = 3072; // (change to 4096 on V100)
const int trDropout = 0.3;

class AsrModel : public fl::Container {
 public:
  AsrModel(
      int64_t dModel,
      int64_t nHead,
      int64_t dimFfd,
      double trDropout,
      int64_t numTransformerLayers,
      int64_t nFeature,
      int64_t nLabel) {
    convFrontend_->add(
        std::make_shared<fl::View>(af::dim4(-1, nFeature, 1, 0)));
    const std::vector<int> convOutDims = {128, 256, 512};
    const std::vector<int> convStrides = {2, 2, 1};
    const std::vector<int> convKws = {3, 3, 1};
    // Time x nFeature x 1 x Batch
    for (int i = 0; i < convOutDims.size(); i++) {
      int convInDim = (i == 0 ? 1 : convOutDims[i - 1]);
      convFrontend_->add(std::make_shared<fl::Conv2D>(
          convInDim,
          convOutDims[i],
          convKws[i],
          convKws[i],
          convStrides[i],
          convStrides[i],
          -1,
          -1,
          1,
          1));
      convFrontend_->add(std::make_shared<fl::BatchNorm>(2, convOutDims[i]));
      convFrontend_->add(std::make_shared<fl::ReLU>());
      convFrontend_->add(std::make_shared<fl::Dropout>(0.1));
    }
    convFrontend_->add(std::make_shared<fl::Reorder>(1, 2, 0, 3));
    // nFeature x channels x Time x Batch
    add(convFrontend_);
    frontlinear_ = std::make_shared<fl::Linear>(10240, dModel);
    add(frontlinear_);
    sinpos_ = std::make_shared<fl::SinusoidalPositionEmbedding>(dModel, 1.0);
    add(sinpos_);
    for (int trIdx = 0; trIdx < numTransformerLayers; trIdx++) {
      auto layer = std::make_shared<fl::Transformer>(
          dModel, dModel / nHead, dimFfd, nHead, 0, trDropout, 0.0, false, false);
      transformers_.push_back(layer);
      add(layer);
    }
    linear_ = std::make_shared<fl::Linear>(dModel, nLabel);
    add(linear_);
  }
  std::vector<fl::Variable> forward(
      const std::vector<fl::Variable>& input) override {
    auto out = input.at(0);
    //    std::cout << out.dims() << std::endl;
    auto xSizes = input.at(1).array();
    // expected input dims T x C x 1 x B
    int T = out.dims(0), B = out.dims(3);
    auto inputMaxSize = af::tile(af::max(xSizes), 1, B);
    af::array inputNotPaddedSize = af::ceil(xSizes * T / inputMaxSize);
    auto padMask = af::iota(af::dim4(T, 1), af::dim4(1, B)) <
        af::tile(inputNotPaddedSize, T, 1);
    out = convFrontend_->forward({out});
    //    std::cout << out.dims() << std::endl;
    out = fl::moddims(
        out, af::dim4(out.dims(0) * out.dims(1), out.dims(2), out.dims(3), 1));
    //    std::cout << out.dims() << std::endl;
    out = frontlinear_->forward(out);
    out = sinpos_->forward({out}).front();
    for (int trIdx = 0; trIdx < transformers_.size(); trIdx++) {
      out = transformers_[trIdx]->forward({out, fl::noGrad(padMask)}).front();
    }
    out = linear_->forward(out);
    return {out.as(input[0].type())};
  }

  std::string prettyString() const override {
    std::ostringstream ss;
    ss << "Model: ";
    ss << convFrontend_->prettyString() << "\n";
    ss << frontlinear_->prettyString() << "\n";
    ss << sinpos_->prettyString() << "\n";
    for (int trIdx = 0; trIdx < transformers_.size(); trIdx++) {
      ss << transformers_[trIdx]->prettyString() << "\n";
    }
    ss << linear_->prettyString() << "\n";
    return ss.str();
  }

 private:
  AsrModel() = default;

  std::shared_ptr<fl::Sequential> convFrontend_{
      std::make_shared<fl::Sequential>()};
  std::vector<std::shared_ptr<fl::Transformer>> transformers_;
  std::shared_ptr<fl::Linear> linear_;
  std::shared_ptr<fl::Linear> frontlinear_;
  std::shared_ptr<fl::SinusoidalPositionEmbedding> sinpos_;

  FL_SAVE_LOAD_WITH_BASE(fl::Container, convFrontend_, transformers_, linear_)
};

CEREAL_REGISTER_TYPE(AsrModel)

std::pair<af::array, af::array> getBatch(int idx) {
  auto inputsize = inputsizes.at(idx);
  auto targetsize = targetsizes.at(idx);
  int T = inputsize.at(0);
  int L = targetsize.at(0);
  af::array x = af::randn(af::dim4(T, FLAGS_num_flt_bank, 1, FLAGS_batch_size));
  af::array y = af::randu(af::dim4(L, FLAGS_batch_size));
  y *= FLAGS_num_labels;
  y = af::max(y.as(s32), (double)(FLAGS_num_labels - 1));
  // std::cout << x.dims() << x.type() << std::endl;
  // std::cout << y.dims() << y.type() << std::endl;
  return std::make_pair(x, y);
}

int main(int argc, char** argv) {
  bool distributed = false;
  fl::init();
  /* Parse or load persistent states */
  gflags::ParseCommandLineFlags(&argc, &argv, false);

  if (FLAGS_distributed_enable) {
    fl::ext::initDistributed(
        FLAGS_distributed_world_rank,
        FLAGS_distributed_world_size,
        FLAGS_distributed_max_devices_per_node,
        FLAGS_distributed_rndv_filepath);
  }

  unsigned batchSize = FLAGS_batch_size;
  unsigned numTransformerLayers = FLAGS_num_transformer_layers;
  std::cout << "---- Starting with batch size " << batchSize << std::endl;
  std::cout << "---- Starting with number of transformer layers "
            << numTransformerLayers << std::endl;

  // ]-------- Create model
  auto model = std::make_shared<AsrModel>(
      dModel,
      nHead,
      dimFeedforward,
      trDropout,
      FLAGS_num_transformer_layers,
      FLAGS_num_flt_bank,
      FLAGS_num_labels);

  std::cout << "---- Created model with " << numTotalParams(model)
            << " parameters." << std::endl;

  if (FLAGS_distributed_enable) {
    fl::distributeModuleGrads(model);
  }

  // ]-------- Optimizer
  auto optimizer = std::make_shared<fl::SGDOptimizer>(
      model->params(), 0.01, 0.9, false, false);

  // ]-------- Loss
  const int kPadIdx_ = 0;
  auto criterion = std::make_shared<
      fl::app::asr::ConnectionistTemporalClassificationCriterion>();

  // ]-------- Benchmark stuff
  bool warmup = true;
  fl::TimeMeter benchmarkMeter{true};
  std::vector<float> times;
  std::chrono::time_point<std::chrono::system_clock> start;
  unsigned numIters = 0;

  model->train();
  criterion->train();

  // ]-------- Training loop
  while (true) {
    auto inlab = getBatch(numIters % 100);
    Variable input = fl::noGrad(inlab.first);
    Variable labels = fl::noGrad(inlab.second);

    // ]--- benchmark things
    if (!warmup && numIters >= numItersPerSample) {
      auto diff = std::chrono::system_clock::now() - start;
      auto elapsed = diff.count() / 1e9;

      std::cout << "benchmark -- Adding sample " << elapsed << " after "
                << numIters << " iterations" << std::endl;
      times.push_back(elapsed);
      numIters = 0;
      start = std::chrono::system_clock::now();
    }
    numIters++;
    if (warmup && numIters > numWarmupIters) {
      std::cout << "benchmark -- Warmup complete" << std::endl;
      warmup = false;
      numIters = 0;
      start = std::chrono::system_clock::now();
    }
    if (times.size() > numSamplesToBench) {
      std::cout << "benchmark -- Benchmarking completed" << std::endl;
      af::array _times = af::array(times.size(), times.data());
      std::cout << "benchmark -- Average iteration time is "
                << std::setprecision(10) << af::mean<double>(_times)
                << " with variance " << af::var<double>(_times) << std::endl;
      return 0;
    }

    // ]---- Forward and backward
    // forward
    auto duration =
        af::constant(input.dims(0), af::dim4(1, input.dims(3)), s32);
    auto output = model->forward({input, fl::noGrad(duration)}).front();
    auto loss = criterion->forward({output, labels}).front();

    optimizer->zeroGrad();

    // backward
    loss.backward();
    optimizer->step();
  }

  return 0;
}
