/*
 * Copyright (c) Facebook, Inc. and its affiliates.
 *
 * This source code is licensed under the BSD-style license found in the
 * LICENSE file in the root directory of this source tree.
 */

#include <chrono>
#include <cmath>
#include <cstdlib>

#include <gflags/gflags.h>
#include <glog/logging.h>

#include "flashlight/app/common/Runtime.h"
#include "flashlight/ext/common/DistributedUtils.h"
#include "flashlight/fl/contrib/contrib.h"

DEFINE_bool(distributed_enable, true, "Enable distributed training");
DEFINE_int64(
    distributed_max_devices_per_node,
    8,
    "the maximum number of devices per training node");
DEFINE_int64(
    distributed_world_rank,
    0,
    "rank of the process (Used if distributed_rndv_filepath is not empty)");
DEFINE_int64(
    distributed_world_size,
    1,
    "total number of the process (Used if distributed_rndv_filepath is not empty)");
DEFINE_string(
    distributed_rndv_filepath,
    "",
    "Shared file path used for setting up rendezvous."
    "If empty, uses MPI to initialize.");
DEFINE_int64(batch_size, 4, "Batch size");
DEFINE_int64(num_transformer_layers, 16, "Number of transformer layers");

using namespace fl;

// ]--- Benchmark parameters
const unsigned numWarmupIters = 100;
const unsigned numItersPerSample = 100;
const unsigned numSamplesToBench = 5;

// Model parameters
const int dModel = 1024; // context size (change to 1024 on V100)
const int nhead = 8; // number of attention heads
const int dimFeedforward = 4096; // (change to 4096 on V100)
const int trDropout = 0.1;
// Number of classes in the dataset
const int numTokens = 100000; // 100k on V100, per CE

// Batch size for training (change depending on how much memory you have)
const int maxTokens =
    1024; // max number of tokens in the batch (i.e. dynamic batching)

class LmModel : public fl::Container {
 public:
  LmModel(
      int64_t dModel,
      int64_t nhead,
      int64_t dimFeedforward = 2048,
      float dropout = 0.1,
      int64_t numTokens = 100000,
      int64_t numTransformerLayers = 16) {
    // Time x B x 1 x 1
    frontend_ = std::make_shared<fl::Sequential>();
    frontend_->add(std::make_shared<fl::Embedding>(dModel, numTokens + 1));
    // nFeature x Time x B x 1
    frontend_->add(std::make_shared<fl::SinusoidalPositionEmbedding>(
        dModel, std::sqrt(dModel)));
    frontend_->add(std::make_shared<fl::Dropout>(0.1));
    // nFeature x Time x Batch x 1
    add(frontend_);
    std::cout << "----- metadata: frontend params: "
              << numTotalParams(frontend_) << std::endl;
    for (int trIdx = 0; trIdx < numTransformerLayers; trIdx++) {
      auto layer = std::make_shared<fl::Transformer>(
          dModel,
          dModel / nhead,
          dimFeedforward,
          nhead,
          0,
          dropout,
          0 /* layerDrop */,
          false,
          false);
      std::cout << "----- metadata: transformer layer " << trIdx
                << " params: " << numTotalParams(layer) << std::endl;
      transformers_.push_back(layer);
      add(layer);
    }

    // Map from feedforward output to num tokens
    finalLinear_ = std::make_shared<Linear>(dModel, numTokens + 1);
    add(finalLinear_);
    std::cout << "----- metadata: final linear params: "
              << numTotalParams(finalLinear_) << std::endl;
  }

  std::vector<fl::Variable> forward(
      const std::vector<fl::Variable>& input) override {
    auto out = input[0];
    // Avoid fp16 usage in any embedding-ralated calls.
    out = frontend_->forward(out);

    // Run all transformer forward passes in fp16
    // out = out.as(f16); // run in fp32
    for (int trIdx = 0; trIdx < transformers_.size(); trIdx++) {
      out = transformers_[trIdx]->forward({out, fl::Variable()}).front();
    }

    // from dModel --> numTokens + 1
    out = finalLinear_->forward(out);

    // Make sure passing fp32 tensor to criterion.
    // Avoid fp16 usage in any embedding-ralated calls.
    return {out.as(f32)};
  }

  std::string prettyString() const override {
    std::ostringstream ss;
    ss << "LmModel: ";
    ss << frontend_->prettyString() << "\n";
    for (int trIdx = 0; trIdx < transformers_.size(); trIdx++) {
      ss << transformers_[trIdx]->prettyString() << "\n";
    }
    return ss.str();
  }

 private:
  LmModel() = default;

  std::shared_ptr<fl::Sequential> frontend_;
  std::vector<std::shared_ptr<fl::Transformer>> transformers_;
  std::shared_ptr<fl::Linear> finalLinear_;
};

af::array getBatch(unsigned outSize, unsigned batchSize) {
  return af::moddims(
      (af::randu(outSize) * numTokens + 1).as(af::dtype::u32),
      af::dim4(outSize / batchSize, batchSize));
}

int main(int argc, char** argv) {
  bool distributed = false;
  fl::init();
  /* Parse or load persistent states */
  gflags::ParseCommandLineFlags(&argc, &argv, false);

  if (FLAGS_distributed_enable) {
    fl::ext::initDistributed(
        FLAGS_distributed_world_rank,
        FLAGS_distributed_world_size,
        FLAGS_distributed_max_devices_per_node,
        FLAGS_distributed_rndv_filepath);
  }

  unsigned batchSize = FLAGS_batch_size;
  unsigned numTransformerLayers = FLAGS_num_transformer_layers;
  std::cout << "---- Starting with batch size " << batchSize << std::endl;
  std::cout << "---- Starting with number of transformer layers "
            << numTransformerLayers << std::endl;

  // ]-------- Create model
  auto model = std::make_shared<LmModel>(
      dModel,
      nhead,
      dimFeedforward,
      trDropout,
      numTokens,
      numTransformerLayers);

  std::cout << "---- Created model with " << numTotalParams(model)
            << " parameters." << std::endl;

  if (FLAGS_distributed_enable) {
    fl::distributeModuleGrads(model);
  }

  // ]-------- Optimizer
  auto optimizer = std::make_shared<fl::SGDOptimizer>(
      model->params(), 0.01, 0.9, false, false);

  // ]-------- Loss
  const int kPadIdx_ = 0;
  auto criterion = std::make_shared<fl::CategoricalCrossEntropy>(
      fl::ReduceMode::NONE, kPadIdx_);

  // ]-------- Benchmark stuff
  bool warmup = true;
  fl::TimeMeter benchmarkMeter{true};
  std::vector<float> times;
  std::chrono::time_point<std::chrono::system_clock> start;
  unsigned numIters = 0;

  model->train();
  criterion->train();

  Variable input = fl::noGrad(getBatch(maxTokens, batchSize));
  Variable labels = fl::noGrad(af::flat(getBatch(maxTokens, batchSize)));

  // ]-------- Training loop
  while (true) {
    // ]--- benchmark things
    if (!warmup && numIters >= numItersPerSample) {
      auto diff = std::chrono::system_clock::now() - start;
      auto elapsed = diff.count() / 1e9;

      std::cout << "benchmark -- Adding sample " << elapsed << " after "
                << numIters << " iterations" << std::endl;
      times.push_back(elapsed);
      numIters = 0;
      start = std::chrono::system_clock::now();
    }
    numIters++;
    if (warmup && numIters > numWarmupIters) {
      std::cout << "benchmark -- Warmup complete" << std::endl;
      warmup = false;
      numIters = 0;
      start = std::chrono::system_clock::now();
    }
    if (times.size() > numSamplesToBench) {
      std::cout << "benchmark -- Benchmarking completed" << std::endl;
      af::array _times = af::array(times.size(), times.data());
      std::cout << "benchmark -- Average iteration time is "
                << std::setprecision(10) << af::mean<double>(_times)
                << " with variance " << af::var<double>(_times) << std::endl;
      return 0;
    }

    // ]---- Forward and backward
    // forward
    auto output = model->forward({input}).front();
    output = fl::moddims(output, af::dim4(numTokens + 1, -1));
    auto loss = criterion->forward(output, labels);

    optimizer->zeroGrad();

    // backward
    loss.backward();
    optimizer->step();
  }

  return 0;
}
