/*
 * Copyright (c) Facebook, Inc. and its affiliates.
 *
 * This source code is licensed under the BSD-style license found in the
 * LICENSE file in the root directory of this source tree.
 */

#include <exception>
#include <iomanip>

#include <gflags/gflags.h>
#include <glog/logging.h>

#include "flashlight/app/common/Runtime.h"
#include "flashlight/app/imgclass/dataset/Imagenet.h"
#include "flashlight/app/imgclass/examples/Defines.h"
#include "flashlight/ext/amp/DynamicScaler.h"
#include "flashlight/ext/common/DistributedUtils.h"
#include "flashlight/ext/image/af/Transforms.h"
#include "flashlight/ext/image/fl/dataset/DistributedDataset.h"
#include "flashlight/ext/image/fl/models/Resnet.h"
#include "flashlight/ext/image/fl/models/ViT.h"
#include "flashlight/fl/dataset/datasets.h"
#include "flashlight/fl/meter/meters.h"
#include "flashlight/fl/optim/optim.h"
#include "flashlight/lib/common/String.h"
#include "flashlight/lib/common/System.h"

// ]--- Benchmark things
const unsigned numWarmupIters = 100;
const unsigned numItersPerSample = 100;
const unsigned numSamplesToBench = 5;

DEFINE_string(data_dir, "", "Directory of imagenet data");
DEFINE_double(train_lr, 0.1f, "Learning rate");
DEFINE_double(train_momentum, 0.9f, "Momentum");

DEFINE_double(train_wd, 1e-4f, "Weight decay");
DEFINE_uint64(train_epochs, 50, "Number of epochs to train");
DEFINE_bool(distributed_enable, true, "Enable distributed training");
DEFINE_int64(
    distributed_max_devices_per_node,
    8,
    "the maximum number of devices per training node");
DEFINE_int64(
    distributed_world_rank,
    0,
    "rank of the process (Used if distributed_rndv_filepath is not empty)");
DEFINE_int64(
    distributed_world_size,
    1,
    "total number of the process (Used if distributed_rndv_filepath is not empty)");
DEFINE_string(
    distributed_rndv_filepath,
    "",
    "Shared file path used for setting up rendezvous."
    "If empty, uses MPI to initialize.");
DEFINE_uint64(data_batch_size, 256, "Total batch size across all gpus");
DEFINE_string(exp_checkpoint_path, "/tmp/model", "Checkpointing prefix path");
DEFINE_int64(exp_checkpoint_epoch, -1, "Checkpoint epoch to load from");
DEFINE_bool(
    generate_detailed_timing,
    false,
    "Enable if detailed timing for "
    "each training step is desired. Enabling this option may adversely "
    "impact the performance.");

// ViT params
DEFINE_int64(model_layers, 12, "Number of transformer layers");
DEFINE_int64(
    model_hidden_emb_size,
    768,
    "Hidden embedding size of transformer block");
DEFINE_int64(model_mlp_size, 3072, "Mlp size of transformer block");
DEFINE_int64(model_heads, 12, "Number of heads of transformer block");

DEFINE_bool(
    fl_amp_use_mixed_precision,
    false,
    "[train] Use mixed precision for training - scale loss and gradients up and down "
    "by a scale factor that changes over time. If no fl optim mode is "
    "specified with --fl_optim_mode when passing this flag, automatically "
    "sets the optim mode to O1.");
DEFINE_double(
    fl_amp_scale_factor,
    65536.,
    "[train] Starting scale factor to use for loss scaling "
    " with mixed precision training");
DEFINE_uint64(
    fl_amp_scale_factor_update_interval,
    2000,
    "[train] Update interval for adjusting loss scaling in mixed precision training");
DEFINE_double(
    fl_amp_max_scale_factor,
    65536.,
    "[train] Maximum value for the loss scale factor in mixed precision training");
DEFINE_string(
    fl_optim_mode,
    "",
    "[train] Sets the flashlight optimization mode. "
    "Optim modes can be O1, O2, or O3.");
DEFINE_string(model_type, "", "Model type {alexnet, vgg, vit, resnet}");
DEFINE_bool(
    random_data,
    false,
    "Use random data (if set to true) or real data (default)");

using namespace fl;
using fl::ext::image::compose;
using fl::ext::image::ImageTransform;
using namespace fl::app::imgclass;

#define FL_LOG_MASTER(lvl) LOG_IF(lvl, (fl::getWorldRank() == 0))

fl::Variable criterion(const fl::Variable& in, const fl::Variable& target) {
  return categoricalCrossEntropy(logSoftmax(in, 0), target);
}

std::shared_ptr<Sequential> alexnet(unsigned numClasses) {
  auto model = std::make_shared<Sequential>();
  model->add(Conv2D(3, 64, 11, 11, 4, 4, 2, 2)); // 224 -> 55
  model->add(ReLU());
  model->add(Pool2D(3, 3, 2, 2)); // 55 ->  27
  model->add(Conv2D(64, 192, 5, 5, 1, 1, 2, 2)); //  27 -> 27
  model->add(ReLU());
  model->add(Pool2D(3, 3, 2, 2)); //  27 ->  13
  model->add(Conv2D(192, 384, 3, 3, 1, 1, 1, 1)); //  13 ->  13
  model->add(ReLU());
  model->add(Conv2D(384, 256, 3, 3, 1, 1, 1, 1)); //  13 ->  13
  model->add(ReLU());
  model->add(Conv2D(256, 256, 3, 3, 1, 1, 1, 1)); //  13 ->  13
  model->add(ReLU());
  model->add(Pool2D(3, 3, 2, 2)); // 13 -> 6
  // Not needed with 224, 224 inputs
  // model->add(Pool2D(6, 6, 1, 1, 0, 0, PoolingMode::AVERAGE_EXCLUDE_PADDING));
  model->add(View(af::dim4(1, 1, -1, 0)));
  model->add(Reorder(2, 3, 0, 1));
  model->add(Dropout());
  model->add(Linear(256 * 6 * 6, 4096));
  model->add(ReLU());
  model->add(Dropout());
  model->add(Linear(4096, 4096));
  model->add(ReLU());
  model->add(Linear(4096, numClasses));
  return model;
}

std::shared_ptr<Sequential> vgg16(unsigned numClasses) {
  auto model = std::make_shared<Sequential>();

  // Block 1
  model->add(Conv2D(3, 64, 3, 3, 1, 1, PaddingMode::SAME, PaddingMode::SAME));
  model->add(ReLU());
  model->add(Conv2D(64, 64, 3, 3, 1, 1, PaddingMode::SAME, PaddingMode::SAME));
  model->add(ReLU());
  model->add(Pool2D(2, 2, 2, 2));
  // Block 2
  model->add(Conv2D(64, 128, 3, 3, 1, 1, PaddingMode::SAME, PaddingMode::SAME));
  model->add(ReLU());
  model->add(
      Conv2D(128, 128, 3, 3, 1, 1, PaddingMode::SAME, PaddingMode::SAME));
  model->add(ReLU());
  model->add(Pool2D(2, 2, 2, 2));
  // Block 3
  model->add(
      Conv2D(128, 256, 3, 3, 1, 1, PaddingMode::SAME, PaddingMode::SAME));
  model->add(ReLU());
  model->add(
      Conv2D(256, 256, 3, 3, 1, 1, PaddingMode::SAME, PaddingMode::SAME));
  model->add(ReLU());
  model->add(
      Conv2D(256, 256, 3, 3, 1, 1, PaddingMode::SAME, PaddingMode::SAME));
  model->add(ReLU());
  model->add(Pool2D(2, 2, 2, 2));
  // Block 4
  model->add(
      Conv2D(256, 512, 3, 3, 1, 1, PaddingMode::SAME, PaddingMode::SAME));
  model->add(ReLU());
  model->add(
      Conv2D(512, 512, 3, 3, 1, 1, PaddingMode::SAME, PaddingMode::SAME));
  model->add(ReLU());
  model->add(
      Conv2D(512, 512, 3, 3, 1, 1, PaddingMode::SAME, PaddingMode::SAME));
  model->add(ReLU());
  model->add(Pool2D(2, 2, 2, 2));
  // Block 5
  model->add(
      Conv2D(512, 512, 3, 3, 1, 1, PaddingMode::SAME, PaddingMode::SAME));
  model->add(ReLU());
  model->add(
      Conv2D(512, 512, 3, 3, 1, 1, PaddingMode::SAME, PaddingMode::SAME));
  model->add(ReLU());
  model->add(
      Conv2D(512, 512, 3, 3, 1, 1, PaddingMode::SAME, PaddingMode::SAME));
  model->add(ReLU());
  model->add(Pool2D(2, 2, 2, 2));

  // Not needed with 224, 224 inputs
  // model->add(Pool2D(6, 6, 1, 1, 0, 0,
  // PoolingMode::AVERAGE_EXCLUDE_PADDING));

  model->add(View(af::dim4(1, 1, -1, 0)));
  model->add(Reorder(2, 3, 0, 1));
  model->add(Dropout());
  model->add(Linear(512 * 7 * 7, 4096));
  model->add(ReLU());
  model->add(Dropout());
  model->add(Linear(4096, 4096));
  model->add(ReLU());
  model->add(Linear(4096, numClasses));
  return model;
}

// Returns the average loss, top 5 error, and top 1 error
std::tuple<double, double, double> evalLoop(
    std::shared_ptr<Sequential> model,
    Dataset& dataset) {
  AverageValueMeter lossMeter;
  TopKMeter top5Acc(5);
  TopKMeter top1Acc(1);

  // Place the model in eval mode.
  model->eval();
  for (auto& example : dataset) {
    auto inputs = noGrad(example[kImagenetInputIdx]);
    auto output = model->forward(inputs);

    auto target = noGrad(example[kImagenetTargetIdx]);

    // Compute and record the loss.
    auto loss = criterion(output, target);
    lossMeter.add(loss.array().scalar<float>());
    top5Acc.add(output.array(), target.array());
    top1Acc.add(output.array(), target.array());
  }
  model->train();
  fl::ext::syncMeter(lossMeter);
  fl::ext::syncMeter(top5Acc);
  fl::ext::syncMeter(top1Acc);

  double loss = lossMeter.value()[0];
  return std::make_tuple(loss, top5Acc.value(), top1Acc.value());
};

int main(int argc, char** argv) {
  fl::init();
  google::InitGoogleLogging(argv[0]);
  google::InstallFailureSignalHandler();
  gflags::ParseCommandLineFlags(&argc, &argv, true);

  const std::string labelPath = lib::pathsConcat(FLAGS_data_dir, "labels.txt");
  const std::string trainList = lib::pathsConcat(FLAGS_data_dir, "train");
  const std::string valList = lib::pathsConcat(FLAGS_data_dir, "val");

  /////////////////////////
  // Setup distributed training
  ////////////////////////
  std::shared_ptr<fl::Reducer> reducer;
  if (FLAGS_distributed_enable) {
    fl::ext::initDistributed(
        FLAGS_distributed_world_rank,
        FLAGS_distributed_world_size,
        FLAGS_distributed_max_devices_per_node,
        FLAGS_distributed_rndv_filepath);
    reducer = std::make_shared<fl::CoalescingReducer>(
        1.0 / fl::getWorldSize(), true, true);
  }
  af::info();
  const int worldRank = fl::getWorldRank();
  const int worldSize = fl::getWorldSize();
  const bool isMaster = (worldRank == 0);

  fl::setDevice(worldRank % 8);
  std::cout << "worldRank " << worldRank << " deviceId " << worldRank % 8
            << std::endl;
  fl::setSeed(worldSize);
  fl::DynamicBenchmark::setBenchmarkMode(true);

  std::shared_ptr<fl::ext::DynamicScaler> dynamicScaler;
  if (FLAGS_fl_amp_use_mixed_precision) {
    FL_LOG_MASTER(INFO)
        << "Mixed precision training enabled. Will perform loss scaling.";
    auto flOptimLevel = FLAGS_fl_optim_mode.empty()
        ? fl::OptimLevel::DEFAULT
        : fl::OptimMode::toOptimLevel(FLAGS_fl_optim_mode);
    fl::OptimMode::get().setOptimLevel(flOptimLevel);

    dynamicScaler = std::make_shared<fl::ext::DynamicScaler>(
        FLAGS_fl_amp_scale_factor,
        FLAGS_fl_amp_max_scale_factor,
        FLAGS_fl_amp_scale_factor_update_interval);
  }

  //////////////////////////
  //  Create datasets
  /////////////////////////
  // These are the mean and std for each channel of Imagenet
  const int randomResizeMax = 480;
  const int randomResizeMin = 256;
  const int randomCropSize = 224;
  const float horizontalFlipProb = 0.5f;
  // TransformDataset will apply each transform in a vector to the respective
  // af::array. Thus, we need to `compose` all of the transforms so are each
  // applied only to the image
  ImageTransform trainTransforms = compose(
      {// randomly resize shortest side of image between 256 to 480 for
       // scale invariance
       fl::ext::image::randomResizeTransform(randomResizeMin, randomResizeMax),
       fl::ext::image::randomCropTransform(randomCropSize, randomCropSize),
       fl::ext::image::normalizeImage(
           fl::app::image::kImageNetMean, fl::app::image::kImageNetStd),
       // Randomly flip image with probability of 0.5
       fl::ext::image::randomHorizontalFlipTransform(horizontalFlipProb)});
  ImageTransform valTransforms = compose(
      {// Resize shortest side to 256, then take a center crop
       fl::ext::image::resizeTransform(randomResizeMin),
       fl::ext::image::centerCropTransform(randomCropSize),
       fl::ext::image::normalizeImage(
           fl::app::image::kImageNetMean, fl::app::image::kImageNetStd)});

  const int64_t batchSizePerGpu = FLAGS_data_batch_size;
  const int64_t prefetchThreads = 10;
  const int64_t prefetchSize = FLAGS_data_batch_size;
  auto labelMap = getImagenetLabels(labelPath);
  auto trainDataset = fl::ext::image::DistributedDataset(
      imagenetDataset(trainList, labelMap, {trainTransforms}),
      worldRank,
      worldSize,
      batchSizePerGpu,
      1, // train_n_repeatedaug
      prefetchThreads,
      prefetchSize,
      fl::BatchDatasetPolicy::SKIP_LAST);

  auto valDataset = fl::ext::image::DistributedDataset(
      imagenetDataset(valList, labelMap, {valTransforms}),
      worldRank,
      worldSize,
      batchSizePerGpu,
      1, // train_n_repeatedaug
      prefetchThreads,
      prefetchSize,
      fl::BatchDatasetPolicy::INCLUDE_LAST);

  //////////////////////////
  //  Load model and optimizer
  /////////////////////////

  std::shared_ptr<Sequential> model;
  if (FLAGS_model_type.empty()) {
    throw std::invalid_argument(
        "--model_type is required -- "
        "can be {alexnet, resnet, vgg, vit}");
  } else if (FLAGS_model_type == "alexnet") {
    model = alexnet(1000);
  } else if (FLAGS_model_type == "vgg") {
    model = vgg16(1000);
  } else if (FLAGS_model_type == "resnet") {
    model = fl::ext::image::resnet50();
  } else if (FLAGS_model_type == "vit") {
    model = std::make_shared<Sequential>();
    model->add(std::make_shared<fl::ext::image::ViT>(
        FLAGS_model_layers,
        FLAGS_model_hidden_emb_size,
        FLAGS_model_mlp_size,
        FLAGS_model_heads,
        0 /* dropout */,
        0. /* layerDrop */,
        1000));
  } else {
    throw std::invalid_argument("--model_type invalid");
  }
  // auto
  // model->add(MobileNetV2(1000));

  std::cout << "Training with model with " << numTotalParams(model)
            << std::endl;

  if (FLAGS_distributed_enable) {
    // synchronize parameters of the model so that the parameters in each
    // process is the same
    fl::allReduceParameters(model);

    // Add a hook to synchronize gradients of model parameters as they are
    // computed
    fl::distributeModuleGrads(model, reducer);
  }

  auto modelParams = model->params();
  SGDOptimizer opt(
      modelParams, FLAGS_train_lr, FLAGS_train_momentum, FLAGS_train_wd);

  auto lrScheduler = [&opt](int epoch) {
    // Adjust learning rate every 30 epoch after 30
    if (epoch == 60 || epoch == 90 || epoch == 120) {
      const float newLr = opt.getLr() * 0.1;
      LOG(INFO) << "Setting learning rate to: " << newLr;
      opt.setLr(newLr);
    }
  };

  // Small utility functions to load and save models
  auto saveModel = [&model, &isMaster](int epoch) {
    if (isMaster) {
      std::string modelPath = FLAGS_exp_checkpoint_path + std::to_string(epoch);
      LOG(INFO) << "Saving model to file: " << modelPath;
      fl::save(modelPath, model);
    }
  };

  auto loadModel = [&model](int epoch) {
    std::string modelPath = FLAGS_exp_checkpoint_path + std::to_string(epoch);
    LOG(INFO) << "Loading model from file: " << modelPath;
    fl::load(modelPath, model);
  };
  if (FLAGS_exp_checkpoint_epoch >= 0) {
    loadModel(FLAGS_exp_checkpoint_epoch);
  }

  // The main training loop
  TopKMeter top5Acc(5);
  TopKMeter top1Acc(1);
  AverageValueMeter trainLossMeter;

  fl::TimeMeter sampleTimerMeter{true};
  fl::TimeMeter fwdTimeMeter{true};
  fl::TimeMeter critFwdTimeMeter{true};
  fl::TimeMeter bwdTimeMeter{true};
  fl::TimeMeter optimTimeMeter{true};
  fl::TimeMeter timeMeter{true};

  bool warmup = true;
  fl::TimeMeter benchmarkMeter{true};
  std::vector<float> times;
  std::chrono::time_point<std::chrono::system_clock> start;

  std::cout << "Random data option is " << FLAGS_random_data << std::endl;

  for (int epoch = (FLAGS_exp_checkpoint_epoch + 1); epoch < FLAGS_train_epochs;
       epoch++) {
    // trainDataset.resample(epoch);
    lrScheduler(epoch);

    size_t numIters = 0;

    // Get an iterator over the data
    int idx = 0;
    for (unsigned i = 0; i < 1000000; ++i) {
      Variable inputs, target;
      if (!FLAGS_random_data) {
        // Real data
        auto example = trainDataset.get(i);
        inputs = noGrad(example[kImagenetInputIdx]);
        target = noGrad(example[kImagenetTargetIdx]);
      } else {
        // Random data
        inputs = noGrad(af::randu({224, 224, 3, batchSizePerGpu}));
        target = noGrad(af::constant(0, {batchSizePerGpu}));
      }
      // ]--- benchmark things
      if (!warmup && numIters >= numItersPerSample) {
        auto diff = std::chrono::system_clock::now() - start;
        auto elapsed = diff.count() / 1e9;

        std::cout << "benchmark -- Adding sample " << elapsed << " after "
                  << numIters << " iterations" << std::endl;
        times.push_back(elapsed);
        numIters = 0;
        start = std::chrono::system_clock::now();
      }
      numIters++;
      if (warmup && numIters > numWarmupIters) {
        std::cout << "benchmark -- Warmup complete" << std::endl;
        warmup = false;
        numIters = 0;
        start = std::chrono::system_clock::now();
      }
      if (times.size() > numSamplesToBench) {
        std::cout << "benchmark -- Benchmarking completed" << std::endl;
        af::array _times = af::array(times.size(), times.data());
        std::cout << "benchmark -- Average iteration time is "
                  << std::setprecision(10) << af::mean<double>(_times)
                  << " with variance " << af::var<double>(_times) << std::endl;
        return 0;
      }

      opt.zeroGrad();

      // Forward
      auto output = model->forward(inputs);
      auto loss = criterion(output, target);

      // Backward
      opt.zeroGrad();
      loss.backward();
      if (FLAGS_distributed_enable) {
        reducer->finalize();
      }

      opt.step();
      fl::sync();
    }
  }
  FL_LOG_MASTER(INFO) << "Training complete";
  return 0;
}
