# An LM model that's a simple stack of transformers with a sinusoidal
# # positional embedding. Uses a fixed dictionary size of 200k

from sinpos import SinusoidalPositionalEmbedding
import torch.nn as nn
import math
from collections import OrderedDict

class ASRModel(nn.Module):
    def __init__(self, d_model, nhead, dim_feedforward=2048, dropout=0.1, activation="relu", num_transformer_layers=16):
        super().__init__()
        frontendlayers = OrderedDict()
        frontendlayers["conv1"] = nn.Conv2d(in_channels=1, out_channels=128, kernel_size=3, stride=2, padding=1)
        frontendlayers["bn1"] = nn.BatchNorm2d(128)
        frontendlayers["relu1"] = nn.ReLU()
        frontendlayers["do1"] = nn.Dropout(dropout)
        frontendlayers["conv2"] = nn.Conv2d(in_channels=128, out_channels=256, kernel_size=3, stride=2, padding=1)
        frontendlayers["bn2"] = nn.BatchNorm2d(256)
        frontendlayers["relu2"] = nn.ReLU()
        frontendlayers["do2"] = nn.Dropout(dropout)
        frontendlayers["conv3"] = nn.Conv2d(in_channels=256, out_channels=512, kernel_size=1, stride=1, padding=0)
        frontendlayers["bn3"] = nn.BatchNorm2d(512)
        frontendlayers["relu3"] = nn.ReLU()
        frontendlayers["do3"] = nn.Dropout(dropout)
        self.frontend = nn.Sequential(frontendlayers)

        self.frontendmap = nn.Linear(10240, d_model)

        amtranslayers = OrderedDict()
        for i in range(num_transformer_layers):
            amtranslayers["trans" + str(i)] = nn.TransformerEncoderLayer(d_model=d_model, nhead=4, dim_feedforward=dim_feedforward, dropout=dropout, activation=activation)
        self.amtrans = nn.Sequential(amtranslayers)
        
        self.amtransmap = nn.Linear(d_model, 31)

        self.sinpos = SinusoidalPositionalEmbedding(
            embedding_dim=d_model, padding_idx=0)

        self.lsm = nn.LogSoftmax(2) # TxBxC or BxTxC
        
    def forward(self, input):
        # print(input.size())
        output = self.frontend(input.view(input.size(0), 1, input.size(1), input.size(2)))
        output = output.permute(2, 0, 1, 3) # TxBxCxD
        outputT = output.size(0)
        outputB = output.size(1)
        output = output.reshape(output.size(0)*output.size(1), output.size(2)*output.size(3))        
        output = self.frontendmap(output)
        output = output.view(outputT, outputB, output.size(1))
        # print(output.size())
        # print(self.sinpos(output[:,:,0]).size())
        output = output + self.sinpos(output[:,:,0])
        output = self.amtrans(output)
        outputT = output.size(0)
        outputB = output.size(1)
        output = output.reshape(output.size(0)*output.size(1), output.size(2))
        output = self.amtransmap(output)
        output = output.view(outputT, outputB, 31)
        # output = output.permute(1, 0, 2) # BxTxC
        # output = output.permute(0, 1, 2) # TxBxC
        # print(output.size())
        output = self.lsm(output)
        return output
