import torch.multiprocessing as mp
import torch.distributed as dist
from torch.nn.parallel import DistributedDataParallel as DDP
import torch
import torch.nn as nn
import torch.optim as optim
import numpy as np
import time
import torchinfo
import os
import copy
import random
import argparse

import inputsizes
import targetsizes

from ASRModel import ASRModel
print("PyTorch Version: ", torch.__version__)

# Model parameters
d_model = 768  # context size (change to 1024 on V100)
nhead = 4  # number of attention heads
dim_feedforward = 3072 # (change to 4096 on V100)
tr_dropout = 0.1 
# Number of classes in the dataset
num_transformer_layers = 36

# Batch size for training (change depending on how much memory you have)
num_flt_bank = 80
num_labels = 31

# Number of epochs to train for
num_epochs = 15

# ]---- Benchmark knobs
# Warmup iterations - the number of warmup iterations to use before measuring benchmarks
num_warmup_iters = 300
num_iters_per_sample = 100
num_samples_to_bench = 5
times = []


def get_batch(batch_size, idx):
    """
    Gets the randomly-generated input batch -- equivalently gets targets
    """
    # print(idx)
    T = inputsizes.data[idx][0]
    L = targetsizes.data[idx][0]
    # print(T, L)
    x = torch.randn(batch_size, T, num_flt_bank)
    y = torch.randint(0, num_labels, (batch_size, L))
    # print(x.size(), y.size())
    # Input should be B x T
    return x, y


def train_model(batch_size, model, criterion, optimizer, num_epochs=25):
    for epoch in range(num_epochs):
        print('Epoch {}/{}'.format(epoch, num_epochs - 1))
        print('-' * 10)

        model.train()  # Set model to training mode
        # ]--- benchmark
        num_iters = 0
        start = time.time()
        warmup = True

        # For benchmark purposes, forward a bunch of random inputs
        while True:
            inputs, labels = get_batch(batch_size=batch_size, idx=num_iters % 100)
            inputs = inputs.cuda()
            labels = labels.cuda()
            num_iters += 1

            # ]--- benchmark things
            if not warmup and num_iters > num_iters_per_sample:
                elapsed = time.time() - start
                print(
                    f"benchmark -- Adding sample {elapsed} after {num_iters_per_sample} iterations")
                times.append(elapsed)
                start = time.time()
                num_iters = 0
            if warmup and num_iters > num_warmup_iters:
                print("benchmark -- Warmup complete")
                warmup = False
                num_iters = 0
                start = time.time()
            if len(times) > num_samples_to_bench:
                print("benchmark -- Average iteration time is ",
                      np.sum(times) / len(times), "with variance", np.var(times))
                return model

            # zero the parameter gradients
            optimizer.zero_grad()

            # forward
            # track history if only in train
            with torch.set_grad_enabled(True):
                # Get model outputs and calculate loss
                outputs = model(inputs)
                input_lengths = torch.full(size=(batch_size,), fill_value=outputs.size(0), dtype=torch.long)
                target_lengths = torch.full(size=(batch_size,), fill_value=labels.size(1), dtype=torch.long)
                loss = criterion(outputs, labels, input_lengths, target_lengths)

                _, preds = torch.max(outputs, 1)

                # # backward + optimize
                loss.backward()
                optimizer.step()

    time_elapsed = time.time() - since
    print('Training complete in {:.0f}m {:.0f}s'.format(
        time_elapsed // 60, time_elapsed % 60))
    print('Best val Acc: {:4f}'.format(best_acc))

    return model


def set_parameter_requires_grad(model):
    for param in model.parameters():
        param.requires_grad = True


def initialize_model():
    model = ASRModel(
        d_model=d_model,
        nhead=nhead,
        dim_feedforward=dim_feedforward,
        dropout=tr_dropout,
        num_transformer_layers=num_transformer_layers,
    ).cuda()
    set_parameter_requires_grad(model)

    return model


def main(args):
    distributed = args.world_size > 1
    gpu = args.rank % torch.cuda.device_count()
    # Initialize distributed
    if distributed:
        print(f"Starting with distributed with rank {args.rank} and gpu {gpu} and world_size {args.world_size} and rndv_path {args.rndv_path}")
        torch.cuda.set_device(gpu)
        # os.environ['RANK'] = str(rank)
        # os.environ['WORLD_SIZE'] = str(world_size)
        # dist.init_process_group("gloo", rank=rank, world_size=world_size)
        torch.distributed.init_process_group(
            backend='nccl',
            rank=args.rank,
            world_size=args.world_size,
            init_method=f"file://{args.rndv_path}",
        )

    # Initialize the model for this run
    model_ft = initialize_model()

    # Send the model to GPU
    if distributed:
        model_ft.cuda(gpu)
        model_ft = torch.nn.parallel.DistributedDataParallel(
            model_ft, device_ids=[gpu])
    else:
        model_ft.cuda()

    # Print the model we just instantiated
    print(model_ft)
    torchinfo.summary(model_ft)
    print("Number params per sum: ", sum(p.numel() for p in model_ft.parameters()))
    print("Trainable params per sum: ", sum(p.numel() for p in model_ft.parameters() if p.requires_grad))


    # Gather the parameters to be optimized/updated in this run. If we are
    #  finetuning we will be updating all parameters. However, if we are
    #  doing feature extract method, we will only update the parameters
    #  that we have just initialized, i.e. the parameters with requires_grad
    #  is True.
    params_to_update = model_ft.parameters()

    # Observe that all parameters are being optimized
    optimizer = optim.SGD(params_to_update, lr=0.001, momentum=0.9)

    # Setup the loss fxn
    criterion = torch.nn.CTCLoss()

    # Train and evaluate
    model_ft = train_model(
        batch_size=args.batch_size,
        model=model_ft,
        criterion=criterion,
        optimizer=optimizer,
        num_epochs=num_epochs,
    )


def start_with_mp(args):
    """
    Used for testing to investigate performance disparities
    """
    def start_with_mp_func(rank, args):
        args.rank = args.rank * args.node_size + rank
        main(args)

    distributed = True
    if distributed:
        # Multi-GPU
        # os.environ['MASTER_ADDR'] = 'localhost'
        # os.environ['MASTER_PORT'] = '12355'
        mp.spawn(
            start_with_mp_func,
            args=(args, ),
            nprocs=args.node_size,
            join=True,
        )
    else:
        main(1, 1)  # single GPU

if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        "PT Vision benchmarks - Proc-based distributed", parents=[]
    )
    parser.add_argument('--rank', default=0, type=int,
                        help='process rank')
    parser.add_argument('--world_size', default=1, type=int,
                        help='number of distributed processes')
    parser.add_argument('--model_name', default='alexnet', type=str,
                        help='Model type [alexnet, vgg, resnet]')
    parser.add_argument('--batch_size', default=32, type=int,
                        help='Batch size')
    parser.add_argument('--node_size', default=8, type=int,
                        help='Node size')
    parser.add_argument('--random_data', action='store_true')
    parser.add_argument('--rndv_path', type=str,
                        help='Rendezvous path')
    args = parser.parse_args()

    print("Running process with rank", args.rank, flush=True)

    main(args);
    # start_with_mp(args)
