import torch.multiprocessing as mp
import torch.distributed as dist
from torch.nn.parallel import DistributedDataParallel as DDP
import torch
import torch.nn as nn
import torch.optim as optim
import numpy as np
import torchvision
from torchvision import datasets, models, transforms
import matplotlib.pyplot as plt
from torch.utils.data import Dataset
import time
import torchinfo
import os
import argparse
import copy
from timm.models.vision_transformer import VisionTransformer
print("PyTorch Version: ", torch.__version__)
print("Torchvision Version: ", torchvision.__version__)

# Distributed

# Top level data directory. Here we assume the format of the directory conforms
#   to the ImageFolder structure
data_dir = "/datasets01/imagenet_resized_144px/060718/061417"

# Number of classes in the dataset
num_classes = 1000

# Number of epochs to train for
num_epochs = 15

# Flag for feature extracting. When False, we finetune the whole model,
#   when True we only update the reshaped layer params
feature_extract = False

# ]---- Benchmark knobs
# Warmup iterations - the number of warmup iterations to use before measuring benchmarks
num_warmup_iters = 100
num_iters_per_sample = 100
num_samples_to_bench = 5
times = []

# Issue is the input shapes are different -- compare PT and TF -- 3, 224, 224 versus 224, 224, 3

def train_model(model, dataloaders, criterion, optimizer, num_epochs, batch_size, device, random_data):
    running_loss = 0.0
    running_corrects = 0
    
    for epoch in range(num_epochs):
        print('Epoch {}/{}'.format(epoch, num_epochs - 1))
        print('-' * 10)

        # Each epoch has a training and validation phase
        for phase in ['train', 'val']:
            if phase == 'train':
                model.train()  # Set model to training mode
            else:
                model.eval()   # Set model to evaluate mode

            # ]--- benchmark
            num_iters = 0
            start = time.time()
            warmup = True

            dataset = iter(dataloaders[phase]) if not random_data else None

            # Iterate over data.
            while True:
                if not random_data:
                    inputs, labels = next(dataset)
                    inputs = inputs.cuda()
                    labels = labels.cuda()
                else:
                    inputs = torch.rand((batch_size, 3, 224, 224), dtype=torch.float32, device=device)
                    labels = torch.randint(0, num_classes, size=(batch_size, ), device=device)
                
                # ]--- benchmark things
                if not warmup and num_iters > num_iters_per_sample:
                    elapsed = time.time() - start
                    print(
                        f"benchmark -- Adding sample {elapsed} after {num_iters_per_sample} iterations", flush=True)
                    times.append(elapsed)
                    start = time.time()
                    num_iters = 0
                num_iters += 1
                if warmup and num_iters > num_warmup_iters:
                    print("benchmark -- Warmup complete", flush=True)
                    warmup = False
                    num_iters = 0
                    start = time.time()
                if len(times) > num_samples_to_bench:
                    print("S")
                    print("benchmark -- Average iteration time is ",
                          np.sum(times) / len(times), "with variance", np.var(times), flush=True)
                    return model

                # zero the parameter gradients
                optimizer.zero_grad()

                # forward
                # track history if only in train
                with torch.set_grad_enabled(phase == 'train'):
                    with torch.cuda.amp.autocast():
                        outputs = model(inputs)
                        loss = criterion(outputs, labels)

                        loss.backward()
                        optimizer.step()

                #     _, preds = torch.max(outputs, 1)

                # running_loss += loss.item() * inputs.size(0)
                # running_corrects += torch.sum(preds == labels.data)

    return model


def set_parameter_requires_grad(model, feature_extracting):
    if feature_extracting:
        for param in model.parameters():
            param.requires_grad = False


def initialize_model(model_name, num_classes, feature_extract, use_pretrained=True):
    # Initialize these variables which will be set in this if statement. Each of these
    #   variables is model specific.
    model_ft = None
    input_size = 0

    if model_name == "resnet":
        """ Resnet50
        """
        model_ft = models.resnet50(pretrained=use_pretrained)
        set_parameter_requires_grad(model_ft, feature_extract)
        num_ftrs = model_ft.fc.in_features
        model_ft.fc = nn.Linear(num_ftrs, num_classes)
        input_size = 224

    elif model_name == "alexnet":
        """ Alexnet
        """
        model_ft = models.alexnet(pretrained=use_pretrained)
        set_parameter_requires_grad(model_ft, feature_extract)
        num_ftrs = model_ft.classifier[6].in_features
        model_ft.classifier[6] = nn.Linear(num_ftrs, num_classes)
        input_size = 224

    elif model_name == "vgg":
        """ VGG16
        """
        model_ft = models.vgg16(pretrained=use_pretrained)
        # model_ft = models.vgg16_bn(pretrained=use_pretrained)
        set_parameter_requires_grad(model_ft, feature_extract)
        num_ftrs = model_ft.classifier[6].in_features
        model_ft.classifier[6] = nn.Linear(num_ftrs, num_classes)
        input_size = 224

    elif model_name == "vit":
        """
        ViT
        """
        model_ft = VisionTransformer()
        set_parameter_requires_grad(model_ft, feature_extract)
        input_size = 224
        
    else:
        print("Invalid model name, exiting...")
        exit()

    return model_ft, input_size


def main(args):
    distributed = args.world_size > 1
    gpu = args.rank % torch.cuda.device_count()
    # Initialize distributed
    if distributed:
        print(f"Starting with distributed with rank {args.rank} and gpu {gpu} and world_size {args.world_size} and rndv_path {args.rndv_path}")
        torch.cuda.set_device(gpu)
        # os.environ['RANK'] = str(args.rank)
        # os.environ['WORLD_SIZE'] = str(args.world_size)
        torch.distributed.init_process_group(
            backend='nccl',
            rank=args.rank,
            world_size=args.world_size,
            init_method=f"file://{args.rndv_path}",
        )

    # Initialize the model for this run
    model_ft, input_size = initialize_model(
        args.model_name, num_classes, feature_extract, use_pretrained=True)

    # Send the model to GPU
    if distributed:
        model_ft.cuda(gpu)
        model_ft = torch.nn.parallel.DistributedDataParallel(
            model_ft, device_ids=[gpu])
    else:
        model_ft.cuda()

    # Print the model we just instantiated
    print(model_ft)

    # Data augmentation and normalization for training
    # Just normalization for validation
    data_transforms = {
        'train': transforms.Compose([
            transforms.RandomResizedCrop(input_size),
            transforms.RandomHorizontalFlip(),
            transforms.ToTensor(),
            transforms.Normalize([0.485, 0.456, 0.406], [0.229, 0.224, 0.225])
        ]),
        'val': transforms.Compose([
            transforms.Resize(input_size),
            transforms.CenterCrop(input_size),
            transforms.ToTensor(),
            transforms.Normalize([0.485, 0.456, 0.406], [0.229, 0.224, 0.225])
        ]),
    }

    print(f"Initializing Datasets and Dataloaders... random data set to {args.random_data}", flush=True)

    # Create training and validation datasets
    dataloaders_dict = None
    if not args.random_data:
        image_datasets = {x:
                          datasets.ImageFolder(os.path.join(data_dir, x), data_transforms[x])
                          for x in ['train', 'val']}
    
        # Create training and validation dataloaders
        if distributed:
            datasamplers_dict = {x: torch.utils.data.distributed.DistributedSampler(
                image_datasets[x],
                shuffle=True,
                num_replicas=args.world_size,
                rank=args.rank) for x in ['train', 'val']}

        dataloaders_dict = {x: torch.utils.data.DataLoader(
            image_datasets[x],
            batch_size=args.batch_size,
            num_workers=(10 if distributed else args.batch_size),
            sampler=(datasamplers_dict[x] if distributed else None)) for x in ['train', 'val']}

    torchinfo.summary(model_ft)
    print("Number params per sum: ", sum(p.numel() for p in model_ft.parameters()))
    print("Trainable params per sum: ", sum(p.numel() for p in model_ft.parameters() if p.requires_grad))


    # Gather the parameters to be optimized/updated in this run. If we are
    #  finetuning we will be updating all parameters. However, if we are
    #  doing feature extract method, we will only update the parameters
    #  that we have just initialized, i.e. the parameters with requires_grad
    #  is True.
    params_to_update = model_ft.parameters()
    print("Params to learn:")
    if feature_extract:
        params_to_update = []
        for name, param in model_ft.named_parameters():
            if param.requires_grad == True:
                params_to_update.append(param)
                print("\t", name)
    else:
        for name, param in model_ft.named_parameters():
            if param.requires_grad == True:
                print("\t", name)

    # Observe that all parameters are being optimized
    optimizer_ft = optim.SGD(params_to_update, lr=0.001, momentum=0.9)

    # Setup the loss fxn
    criterion = nn.CrossEntropyLoss().cuda(gpu)  # -- includes softmax

    # Train and evaluate
    model_ft = train_model(
        model=model_ft,
        dataloaders=dataloaders_dict,
        criterion=criterion,
        optimizer=optimizer_ft,
        num_epochs=num_epochs,
        batch_size=args.batch_size,
        device=gpu,
        random_data=args.random_data,
    )


def start_with_mp(args):
    """
    Used for testing to investigate performance disparities
    """
    def start_with_mp_func(rank, args):
        args.rank = args.rank * args.node_size + rank
        main(args)

    distributed = True    
    if distributed:
        # Multi-GPU        
        # os.environ['MASTER_ADDR'] = 'localhost'
        # os.environ['MASTER_PORT'] = '12355'
        mp.spawn(
            start_with_mp_func,
            args=(args, ),
            nprocs=args.node_size,
            join=True,
        )
    else:
        main(1, 1)  # single GPU
    
if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        "PT Vision benchmarks - Proc-based distributed", parents=[]
    )
    parser.add_argument('--rank', default=0, type=int,
                        help='process rank')
    parser.add_argument('--world_size', default=1, type=int,
                        help='number of distributed processes')
    parser.add_argument('--model_name', default='alexnet', type=str,
                        help='Model type [alexnet, vgg, resnet]')
    parser.add_argument('--batch_size', default=32, type=int,
                        help='Batch size')
    parser.add_argument('--node_size', default=8, type=int,
                        help='Node size')
    parser.add_argument('--random_data', action='store_true')
    parser.add_argument('--rndv_path', type=str,
                        help='Rendezvous path')
    args = parser.parse_args()

    print("Running process with rank", args.rank, flush=True)

    main(args);
    # start_with_mp(args)
