import tensorflow as tf
from tensorflow import keras
from tensorflow.keras import layers, models, regularizers
import numpy as np
import time
import transformer
import argparse
import json
import os

import inputsizes
import targetsizes

# ]--- metadata

# Model parameters
d_model = 768  # context size (change to 1024 on V100)
nhead = 4  # number of attention heads
dim_feedforward = 3072  # (change to 4096 on V100)
tr_dropout = 0.3
# Number of classes in the dataset
num_transformer_layers = 36

# Batch size for training (change depending on how much memory you have)
batch_size = 10
num_flt_bank = 80
num_labels = 31

# Number of epochs to train for
num_epochs = 15

# ]---- Benchmark knobs
# Warmup iterations - the number of warmup iterations to use before measuring benchmarks
num_warmup_iters = 100
num_iters_per_sample = 100
num_samples_to_bench = 5
times = []
thread_pool_size = 10

### ---------- ###

maxT = 0
for i in range(len(inputsizes.data)):
    maxT = max(maxT, inputsizes.data[i][0])
maxL = 0
for i in range(len(targetsizes.data)):
    maxL = max(maxL, targetsizes.data[i][0])

print('maxT:', maxT)
print('maxL:', maxL)

def get_batch(batch_size, idx):
    """
    Gets the randomly-generated input batch -- equivalently gets targets
    """
    # print(idx)
    T = inputsizes.data[idx][0]
    L = targetsizes.data[idx][0]
    # print(T, L)
    # (batch_size, height, width, channels)
    x = tf.random.uniform((batch_size, T, num_flt_bank, 1))
    y = tf.random.uniform((batch_size, L), minval=0, maxval=num_labels-1, dtype=tf.int32)
    # print(x.size(), y.size())
    # Input should be B x T
    return x, y


def train(batch_size, distributed, epochs, model, optimizer, loss_fn, train_acc_metric, val_acc_metric, strategy):
    def train_step(inputs, labels):
        with tf.GradientTape() as tape:
            # loss_value = tf.math.reduce_sum(model(inputs, training=True))
            logits = model(inputs, training=True)
            loss_value = loss_fn(y_true=labels,
                                 y_pred=logits)
        grads = tape.gradient(loss_value, model.trainable_weights)
        optimizer.apply_gradients(zip(grads, model.trainable_weights))
        # Update training metric
        # train_acc_metric.update_state(labels, logits)
        return loss_value

    @tf.function(experimental_relax_shapes=True)
    def train_loop(inputs, labels):
        return train_step(inputs, labels)

    @tf.function(experimental_relax_shapes=True)
    def distributed_train_loop(inputs, labels):
        replica_losses = strategy.run(train_step, args=(inputs, labels))
        return strategy.reduce(tf.keras.losses.Reduction.SUM, replica_losses, axis=None)

    for epoch in range(epochs):
        print("\nStart of epoch %d" % (epoch,))

        # ]--- benchmark
        num_iters = 0
        start = time.time()
        warmup = True


        # For benchmark purposes, forward a bunch of random inputs
        while True:
            inputs, labels = get_batch(batch_size=batch_size, idx=num_iters)
            # print(tf.shape(inputs))
            num_iters += 1

            # ]--- benchmark things
            if not warmup and num_iters > num_iters_per_sample:
                elapsed = time.time() - start
                print(
                    f"benchmark -- Adding sample {elapsed} after {num_iters_per_sample} iterations")
                times.append(elapsed)
                start = time.time()
                num_iters = 0
            if warmup and num_iters > num_warmup_iters:
                print("benchmark -- Warmup complete")
                warmup = False
                num_iters = 0
                start = time.time()
            if len(times) > num_samples_to_bench:
                print("benchmark -- Benchmarking completed")
                print("benchmark -- Average iteration time is ",
                      np.sum(times) / len(times), "with variance", np.var(times))
                return

            if distributed:
                distributed_train_loop(inputs, labels)
            else:
                train_loop(inputs, labels)

                
def initialize_model():
    # def backend_reshape1(x):
    #     return tf.keras.backend.reshape(
    #         x,
    #         (10*417, 20*512)
    #     )
    # def backend_reshape2(x):
    #     return tf.keras.backend.reshape(
    #         x,
    #         (10, 417, d_model)
    #     )
    # def backend_reshape3(x):
    #     return tf.keras.backend.reshape(
    #         x,
    #         (10*417, d_model)
    #     )
    # def backend_reshape4(x):
    #     return tf.keras.backend.reshape(
    #         x,
    #         (10, 417, num_labels)
    #     )

    model = tf.keras.Sequential()
    # batch_size, height (T), width (fltbank), channels (1)
    model.add(layers.Conv2D(filters=128, kernel_size=3, strides=2, padding='same'))
    model.add(layers.BatchNormalization(axis=3))
    model.add(layers.ReLU())
    model.add(layers.Dropout(0.1))
    model.add(layers.Conv2D(filters=256, kernel_size=3, strides=2, padding='same'))
    model.add(layers.BatchNormalization(axis=3))
    model.add(layers.ReLU())
    model.add(layers.Dropout(0.1))
    model.add(layers.Conv2D(filters=512, kernel_size=1, strides=1, padding='same'))
    model.add(layers.BatchNormalization(axis=3))
    model.add(layers.ReLU())
    model.add(layers.Dropout(0.1))
    model.add(layers.Reshape((-1, 20*512)))
    # model.add(layers.Lambda(backend_reshape1))
    model.add(layers.Dense(d_model))
    # model.add(layers.Lambda(backend_reshape2))
    for i in range(num_transformer_layers):
        model.add(transformer.Encoder(embed_dim=d_model, num_heads=nhead, feed_forward_dim=dim_feedforward, rate=tr_dropout, maximum_position_encoding=((int(maxT/4)+1) if (i==0) else 0)))
    # model.add(layers.Lambda(backend_reshape3))
    model.add(layers.Dense(num_labels, activation="softmax"))
    # model.add(layers.Lambda(backend_reshape4))
    model.build(input_shape=(None, None, num_flt_bank, 1))
    return model


class CTCLayer(layers.Layer):
    def __init__(self, name=None):
        super().__init__(name=name)
        self.loss_fn = keras.backend.ctc_batch_cost
        
    def call(self, y_true, y_pred):
        # Compute the training-time loss value and add it
        # to the layer using
        batch_len = tf.cast(tf.shape(y_true)[0], dtype="int64")
        input_length = tf.cast(tf.shape(y_pred)[1], dtype="int64")
        label_length = tf.cast(tf.shape(y_true)[1], dtype="int64")
    
        input_length = input_length * tf.ones(shape=(batch_len, 1), dtype="int64")
        label_length = label_length * tf.ones(shape=(batch_len, 1), dtype="int64")

        loss = self.loss_fn(y_true, y_pred, input_length, label_length)
        self.add_loss(loss)
        
        # At test time, just return the computed predictions
        return y_pred

def main(args):
    distributed = args.distributed
    strategy = None
    if distributed:
        communication_options = tf.distribute.experimental.CommunicationOptions(
            implementation=tf.distribute.experimental.CommunicationImplementation.NCCL)
        strategy = tf.distribute.MultiWorkerMirroredStrategy(communication_options=communication_options)

    def run_training(batch_size, distributed, strategy=None):
        print("Initialization -- distributed is ", distributed)

        # Get model
        model = initialize_model()
        print(model.summary())

        # Instantiate an optimizer.
        optimizer = keras.optimizers.SGD(learning_rate=1e-3, momentum=0.9)
        # Instantiate a loss function.
        loss_fn = CTCLayer()

        # Prepare the metrics.
        train_acc_metric = keras.metrics.SparseCategoricalAccuracy()
        val_acc_metric = keras.metrics.SparseCategoricalAccuracy()

        train(
            batch_size=batch_size,
            distributed=distributed,
            epochs=2,
            model=model,
            optimizer=optimizer,
            loss_fn=loss_fn,
            train_acc_metric=train_acc_metric,
            val_acc_metric=val_acc_metric,
            strategy=strategy,
        )

    if distributed:
        with strategy.scope():
            run_training(
                batch_size=args.batch_size,
                distributed=distributed,
                strategy=strategy,
            )
    else:
        tf.config.threading.set_intra_op_parallelism_threads(thread_pool_size)
        tf.config.threading.set_inter_op_parallelism_threads(thread_pool_size)
        run_training(
            batch_size=args.batch_size,
            distributed=distributed,
        )

def get_hosts_list(filename, port=15000):
    with open(filename, 'r') as f:
        lines = f.readlines()
        lines = [f"{line.rstrip()}:{port}" for line in lines]
        return lines

if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        "PT Vision benchmarks - Proc-based distributed", parents=[]
    )
    parser.add_argument('--node_rank', default=0, type=int,
                        help='Node rank')
    parser.add_argument('--hosts_filepath', default=None, type=str,
                        help='file containing learnfair machines on which to run')
    parser.add_argument('--model_name', default='alexnet', type=str,
                        help='Model type [alexnet, vgg, resnet]')
    parser.add_argument('--batch_size', default=128, type=int,
                        help='Batch size')
    parser.add_argument('--random_data', action='store_true')
    args = parser.parse_args()

    if args.hosts_filepath:
        os.environ.pop('TF_CONFIG', None)
        hosts_list = get_hosts_list(args.hosts_filepath)
        print(f"Starting on hosts {hosts_list}")
        tf_config = {
            'cluster': {
                'worker': hosts_list
            },
            'task': {'type': 'worker', 'index': args.node_rank}
        }
        os.environ['TF_CONFIG'] = json.dumps(tf_config)
        nhosts = len(hosts_list)
        print(f"Will run training on {nhosts} machines and {nhosts * 8} gpu")
        args.distributed = True
    else:
        args.distributed = False

    main(args)
