# The PyTorch transformer, but slightly tweaked -- define the decoder
# layer without attending to the encoder (remove multihead attention
# to the encoder and remove bias everywhere). This ensures the
# implementation has parity with Flashlight and TensorFlow

import tensorflow as tf

from sinpos import positional_encoding

from official.nlp.transformer import attention_layer
from official.nlp.transformer import embedding_layer
from official.nlp.transformer import ffn_layer

from tensorflow.keras.layers import Dense
from tensorflow.keras import layers

# revert to postlayernorm

class Encoder(layers.Layer):
    def __init__(self, embed_dim, num_heads, feed_forward_dim, rate=0.1, maximum_position_encoding=0):
        super().__init__()
        self.att = layers.MultiHeadAttention(num_heads=num_heads, key_dim=int(embed_dim/num_heads), use_bias=False)
        self.ffn = tf.keras.Sequential(
            [
                layers.Dense(feed_forward_dim, activation="relu", use_bias=False),
                layers.Dense(embed_dim, use_bias=False),
            ]
        )
        self.layernorm1 = layers.LayerNormalization(epsilon=1e-6)
        self.layernorm2 = layers.LayerNormalization(epsilon=1e-6)
        self.dropout1 = layers.Dropout(rate)
        self.dropout2 = layers.Dropout(rate)
        if maximum_position_encoding > 0:
            self.pos_encoding = positional_encoding(
            maximum_position_encoding, embed_dim)

    def call(self, inputs, training):
        if hasattr(self, 'pos_encoding'):
            seq_len = tf.shape(inputs)[1]        
            inputs += self.pos_encoding[:, :seq_len, :]
            
        attn_output = self.att(inputs, inputs)
        attn_output = self.dropout1(attn_output, training=training)
        out1 = self.layernorm1(inputs + attn_output)
        ffn_output = self.ffn(out1)
        ffn_output = self.dropout2(ffn_output, training=training)
        return self.layernorm2(out1 + ffn_output)

class PrePostProcessingWrapper(tf.keras.layers.Layer):
    """Wrapper class that applies layer pre-processing and post-processing."""

    def __init__(self, layer, params):
        super(PrePostProcessingWrapper, self).__init__()
        self.layer = layer
        self.params = params
        self.postprocess_dropout = params["layer_postprocess_dropout"]

    def build(self, input_shape):
        # Create normalization layer
        self.layer_norm = tf.keras.layers.LayerNormalization(
            epsilon=1e-6, dtype="float32")
        super(PrePostProcessingWrapper, self).build(input_shape)

    def get_config(self):
        return {
            "params": self.params,
        }

    def call(self, x, *args, **kwargs):
        """Calls wrapped layer with same parameters."""
        # Preprocessing: apply layer normalization
        training = kwargs["training"]

        # Get layer output
        y = self.layer(x, *args, **kwargs)

        # Postprocessing: apply dropout and residual connection
        if training:
            y = tf.nn.dropout(y, rate=self.postprocess_dropout)
        return self.layer_norm(x + y)


class DecoderStack(tf.keras.layers.Layer):
    """Transformer decoder stack.
    Like the encoder stack, the decoder stack is made up of N identical layers.
    Each layer is composed of the sublayers:
      1. Self-attention layer
      2. Multi-headed attention layer combining encoder outputs with results from
         the previous self-attention layer.
      3. Feedforward network (2 fully-connected layers)
    """

    def __init__(self, params):
        super(DecoderStack, self).__init__()
        self.params = params
        self.layers = []

    def build(self, input_shape):
        """Builds the decoder stack."""
        params = self.params
        for _ in range(params["num_hidden_layers"]):
            self_attention_layer = attention_layer.SelfAttention(
                params["hidden_size"], params["num_heads"],
                params["attention_dropout"])
            feed_forward_network = ffn_layer.FeedForwardNetwork(
                params["hidden_size"], params["filter_size"], params["relu_dropout"])

            self.layers.append([
                PrePostProcessingWrapper(self_attention_layer, params),
                PrePostProcessingWrapper(feed_forward_network, params)
            ])
        self.output_normalization = tf.keras.layers.LayerNormalization(
            epsilon=1e-6, dtype="float32")
        super(DecoderStack, self).build(input_shape)

    def get_config(self):
        return {
            "params": self.params,
        }

    def call(self,
             decoder_inputs,
             decoder_self_attention_bias,
             training,
             cache=None,
             decode_loop_step=None):
        """Return the output of the decoder layer stacks.
        Args:
          decoder_inputs: A tensor with shape [batch_size, target_length,
            hidden_size].
          decoder_self_attention_bias: A tensor with shape [1, 1, target_len,
            target_length], the bias for decoder self-attention layer.
          training: A bool, whether in training mode or not.
          cache: (Used for fast decoding) A nested dictionary storing previous
            decoder self-attention values. The items are:
              {layer_n: {"k": A tensor with shape [batch_size, i, key_channels],
                         "v": A tensor with shape [batch_size, i, value_channels]},
                           ...}
          decode_loop_step: An integer, the step number of the decoding loop. Used
            only for autoregressive inference on TPU.
        Returns:
          Output of decoder layer stack.
          float32 tensor with shape [batch_size, target_length, hidden_size]
        """
        for n, layer in enumerate(self.layers):
            self_attention_layer = layer[0]
            feed_forward_network = layer[1]

            # Run inputs through the sublayers.
            layer_name = "layer_%d" % n
            layer_cache = cache[layer_name] if cache is not None else None
            with tf.name_scope(layer_name):
                with tf.name_scope("self_attention"):
                    decoder_inputs = self_attention_layer(
                        decoder_inputs,
                        decoder_self_attention_bias,
                        training=training,
                        cache=layer_cache,
                        decode_loop_step=decode_loop_step)
                with tf.name_scope("ffn"):
                    decoder_inputs = feed_forward_network(
                        decoder_inputs, training=training)

        return self.output_normalization(decoder_inputs)


class Decoder(tf.keras.layers.Layer):
    def __init__(self, num_layers, d_model, num_heads, dff, target_vocab_size,
                 maximum_position_encoding, rate=0.1):
        super(Decoder, self).__init__()

        self.d_model = d_model
        self.num_layers = num_layers

        self.embedding = tf.keras.layers.Embedding(target_vocab_size, d_model)
        self.pos_encoding = positional_encoding(
            maximum_position_encoding, d_model)

        self.dec_layers = DecoderStack(
            {
                "hidden_size": d_model,
                "num_heads": num_heads,
                "filter_size": dff,
                "attention_dropout": rate,
                "num_hidden_layers": num_layers,
                "relu_dropout": rate,
                "layer_postprocess_dropout": rate,
            }
        )

        self.dropout = tf.keras.layers.Dropout(rate)

        self.linear = Dense(target_vocab_size + 1, use_bias=True)

    def call(self, x, training):

        seq_len = tf.shape(x)[1]
        
        x = self.embedding(x)  # (batch_size, target_seq_len, d_model)
        
        x *= tf.math.sqrt(tf.cast(self.d_model, tf.float32))
        x += self.pos_encoding[:, :seq_len, :]

        x = self.dropout(x, training=training)

        x = self.dec_layers(x, tf.zeros(1), training)

        x = self.linear(x)

        # x.shape == (batch_size, target_seq_len, d_model)
        return x
