import tensorflow as tf
from tensorflow import keras
from tensorflow.keras import layers, models, regularizers
import numpy as np
import time
import argparse
import json
import os

from transformer import Decoder

# ]--- metadata

# Model parameters
d_model = 1024  # context size (change to 1024 on V100)
nhead = 8  # number of attention heads
dim_feedforward = 4096  # (change to 4096 on V100)
tr_dropout = 0.1
# Number of classes in the dataset
num_tokens = 100000  # 100k on V100, per CE
num_transformer_layers = 16

# Batch size for training (change depending on how much memory you have)
max_tokens = 1024  # max number of tokens in the batch (i.e. dynamic batching)

# Number of epochs to train for
num_epochs = 15

# ]---- Benchmark knobs
# Warmup iterations - the number of warmup iterations to use before measuring benchmarks
num_warmup_iters = 100
num_iters_per_sample = 100
num_samples_to_bench = 5
times = []
thread_pool_size = 10

### ---------- ###


def get_batch(batch_size, size):
    """
    Gets the randomly-generated input batch
    """
    # Input should be B x T
    return tf.random.uniform((batch_size, max_tokens // batch_size), minval=1, maxval=num_tokens, dtype=tf.int32)


def get_labels(size):
    # Labels should be B x T for TF's CE
    return tf.random.uniform((max_tokens, num_tokens + 1), minval=1, maxval=num_tokens, dtype=tf.int32)


def train(batch_size, distributed, epochs, model, optimizer, loss_fn, train_acc_metric, val_acc_metric, strategy):
    def train_step(inputs, labels):
        with tf.GradientTape() as tape:
            # loss_value = tf.math.reduce_sum(model(inputs, training=True))
            logits = tf.reshape(
                model(inputs, training=True), (-1, num_tokens + 1))
            loss_value = loss_fn(y_true=labels,
                                 y_pred=logits)
        grads = tape.gradient(loss_value, model.trainable_weights)
        optimizer.apply_gradients(zip(grads, model.trainable_weights))
        # Update training metric
        # train_acc_metric.update_state(labels, logits)
        return loss_value

    @tf.function
    def train_loop(inputs, labels):
        return train_step(inputs, labels)

    @tf.function
    def distributed_train_loop(inputs, labels):
        replica_losses = strategy.run(train_step, args=(inputs, labels))
        return strategy.reduce(tf.keras.losses.Reduction.SUM, replica_losses, axis=None)

    for epoch in range(epochs):
        print("\nStart of epoch %d" % (epoch,))

        # ]--- benchmark
        num_iters = 0
        start = time.time()
        warmup = True


        # For benchmark purposes, forward a bunch of random inputs
        while True:
            inputs = get_batch(batch_size=batch_size, size=max_tokens)
            labels = tf.stop_gradient(get_batch(batch_size=batch_size, size=max_tokens))

            # ]--- benchmark things
            if not warmup and num_iters > num_iters_per_sample:
                elapsed = time.time() - start
                print(
                    f"benchmark -- Adding sample {elapsed} after {num_iters_per_sample} iterations")
                times.append(elapsed)
                start = time.time()
                num_iters = 0
            num_iters += 1
            if warmup and num_iters > num_warmup_iters:
                print("benchmark -- Warmup complete")
                warmup = False
                num_iters = 0
                start = time.time()
            if len(times) > num_samples_to_bench:
                print("benchmark -- Benchmarking completed")
                print("benchmark -- Average iteration time is ",
                      np.sum(times) / len(times), "with variance", np.var(times))
                return

            if distributed:
                distributed_train_loop(inputs, labels)
            else:
                train_loop(inputs, labels)

def initialize_model(batch_size):
    model = tf.keras.Sequential([
        Decoder(
            d_model=d_model,
            num_heads=nhead,
            dff=dim_feedforward,
            rate=tr_dropout,
            target_vocab_size=num_tokens,
            num_layers=num_transformer_layers,
            maximum_position_encoding=1.0,
        ),
    ],
        name="Transformer",
    )
    model.build(input_shape=(batch_size, max_tokens // batch_size))
    return model

def main(args):
    distributed = args.distributed
    strategy = None
    if distributed:
        communication_options = tf.distribute.experimental.CommunicationOptions(
            implementation=tf.distribute.experimental.CommunicationImplementation.NCCL)
        strategy = tf.distribute.MultiWorkerMirroredStrategy(communication_options=communication_options)
    
    def run_training(batch_size, distributed, strategy=None):
        print("Initialization -- distributed is ", distributed, flush=True)

        # Get model
        model = initialize_model(args.batch_size)
        print(model.summary())

        # Instantiate an optimizer.
        optimizer = keras.optimizers.SGD(learning_rate=1e-3, momentum=0.9)
        # Instantiate a loss function.
        loss_fn = keras.losses.SparseCategoricalCrossentropy(
            from_logits=True, reduction=tf.keras.losses.Reduction.NONE)

        # Prepare the metrics.
        train_acc_metric = keras.metrics.SparseCategoricalAccuracy()
        val_acc_metric = keras.metrics.SparseCategoricalAccuracy()

        train(
            batch_size=batch_size,
            distributed=distributed,
            epochs=2,
            model=model,
            optimizer=optimizer,
            loss_fn=loss_fn,
            train_acc_metric=train_acc_metric,
            val_acc_metric=val_acc_metric,
            strategy=strategy,
        )

    # Distributed training
    if distributed:
        with strategy.scope():
            run_training(
                batch_size=args.batch_size,
                distributed=distributed,
                strategy=strategy,
            )
    else:
        tf.config.threading.set_intra_op_parallelism_threads(thread_pool_size)
        tf.config.threading.set_inter_op_parallelism_threads(thread_pool_size)
        run_training(
            batch_size=args.batch_size,
            distributed=distributed,
        )

def get_hosts_list(filename, port=15000):
    with open(filename, 'r') as f:
        lines = f.readlines()
        lines = [f"{line.rstrip()}:{port}" for line in lines]
        return lines
        
if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        "PT Vision benchmarks - Proc-based distributed", parents=[]
    )
    parser.add_argument('--node_rank', default=0, type=int,
                        help='Node rank')
    parser.add_argument('--hosts_filepath', default=None, type=str,
                        help='file containing learnfair machines on which to run')
    parser.add_argument('--model_name', default='alexnet', type=str,
                        help='Model type [alexnet, vgg, resnet]')
    parser.add_argument('--batch_size', default=128, type=int,
                        help='Batch size')
    parser.add_argument('--random_data', action='store_true')
    args = parser.parse_args()

    if args.hosts_filepath:
        os.environ.pop('TF_CONFIG', None)
        hosts_list = get_hosts_list(args.hosts_filepath)
        print(f"Starting on hosts {hosts_list}")
        tf_config = {
            'cluster': {
                'worker': hosts_list
            },
            'task': {'type': 'worker', 'index': args.node_rank}
        }
        os.environ['TF_CONFIG'] = json.dumps(tf_config)
        nhosts = len(hosts_list)
        print(f"Will run training on {nhosts} machines and {nhosts * 8} gpu")
        args.distributed = True
    else:
        args.distributed = False

    main(args)
