import tensorflow as tf
from tensorflow import keras
from tensorflow.keras import layers, models, regularizers
import numpy as np
import time
import argparse
import json
import os

from alexnet import alexnet
from vit_model import vit

# ]--- metadata

data_root = "/datasets01/imagenet_resized_144px/060718/061417"
num_classes = 1000

# ]---- Benchmark knobs
# Warmup iterations - the number of warmup iterations to use before measuring benchmarks
num_warmup_iters = 100
num_iters_per_sample = 100
num_samples_to_bench = 5
times = []
thread_pool_size = 10

### ---------- ###


def get_datasets(datadir, batch_size):
    img_height = 224
    img_width = 224

    options = tf.data.Options()
    options.experimental_threading.private_threadpool_size = thread_pool_size

    train_ds = tf.keras.preprocessing.image_dataset_from_directory(
        str(datadir) + '/train',
        validation_split=0.2,
        subset="training",
        label_mode='int', # sparse
        seed=123,
        image_size=(img_height, img_width),
        batch_size=batch_size
    ).with_options(options)

    val_ds = tf.keras.preprocessing.image_dataset_from_directory(
        str(datadir) + '/val',
        validation_split=0.2,
        subset="validation",
        label_mode='int', # sparse
        seed=123,
        image_size=(img_height, img_width),
        batch_size=batch_size
    ).with_options(options)

    return train_ds, val_ds


def train(distributed, epochs, model, train_dataset, optimizer, loss_fn, train_acc_metric, val_acc_metric, strategy, random_data, batch_size):
    def train_step(images, labels):
        with tf.GradientTape() as tape:
            logits = model(images, training=True)
            loss_value = loss_fn(y_true=labels, y_pred=logits)
        grads = tape.gradient(loss_value, model.trainable_weights)
        optimizer.apply_gradients(zip(grads, model.trainable_weights))
        # Update training metric
        train_acc_metric.update_state(labels, logits)
        return loss_value

    @tf.function
    def train_loop(images, labels):
        return train_step(images, labels)

    @tf.function
    def distributed_train_loop(images, labels):
        replica_losses = strategy.run(train_step, args=(images, labels))
        return strategy.reduce(tf.keras.losses.Reduction.SUM, replica_losses, axis=None)

    for epoch in range(epochs):
        print("\nStart of epoch %d" % (epoch,))
        print("Random data is set as", random_data)

        # ]--- benchmark
        num_iters = 0
        start = time.time()
        warmup = True

        dataset_iter = iter(enumerate(train_dataset)) if not random_data else None
        
        # Iterate over the batches of the dataset.
        step = 0
        while True:
            if not random_data:
                step, (x_batch_train, y_batch_train) = next(dataset_iter)
            else:
                step = step + 1
                x_batch_train = tf.random.uniform(shape=(batch_size, 224, 224, 3))
                y_batch_train = tf.random.uniform(shape=(batch_size, ), maxval=num_classes, dtype=tf.dtypes.int32)
            
            # ]--- benchmark things
            if not warmup and num_iters > num_iters_per_sample:
                elapsed = time.time() - start
                print(
                    f"benchmark -- Adding sample {elapsed} after {num_iters_per_sample} iterations")
                times.append(elapsed)
                start = time.time()
                num_iters = 0
            num_iters += 1
            if warmup and num_iters > num_warmup_iters:
                print("benchmark -- Warmup complete")
                warmup = False
                num_iters = 0
                start = time.time()
            if len(times) > num_samples_to_bench:
                print("benchmark -- Benchmarking completed")
                print("benchmark -- Average iteration time is ",
                      np.sum(times) / len(times), "with variance", np.var(times))
                return

            if distributed:
                distributed_train_loop(x_batch_train, y_batch_train)
            else:
                train_loop(x_batch_train, y_batch_train)

            # if step % 200 == 0:
            #     print(
            #         "Training loss (for one batch) at step %d: %.4f"
            #         % (step, float(loss_value))
            #     )
            #     print("Seen so far: %d samples" % ((step + 1) * batch_size))

        # Display metrics at the end of each epoch.
        train_acc = train_acc_metric.result()
        print("Training acc over epoch: %.4f" % (float(train_acc),))

        # Reset training metrics at the end of each epoch
        train_acc_metric.reset_states()

        val_acc = val_acc_metric.result()
        val_acc_metric.reset_states()
        print("Validation acc: %.4f" % (float(val_acc),))
        print("Time taken: %.2fs" % (time.time() - start))


def main(args):
    distributed = args.distributed
    strategy = None
    if distributed:
        communication_options = tf.distribute.experimental.CommunicationOptions(
            implementation=tf.distribute.experimental.CommunicationImplementation.NCCL)
        strategy = tf.distribute.MultiWorkerMirroredStrategy(communication_options=communication_options)

    def run_training(batch_size, distributed, strategy=None):
        print("Initialization -- distributed is", distributed)        

        # tf.keras.backend.set_image_data_format('channels_first')
        input_shape = (224, 224, 3)  # (3, 224, 224)

        # Create datasets
        train_dataset, val_ds = None, None
        if not args.random_data:
            train_dataset, val_ds = get_datasets(
                datadir=data_root,
                batch_size=args.batch_size,
            )
            train_dataset = train_dataset.prefetch(10)
            val_ds = val_ds.prefetch(10)
        if distributed:
            train_dataset = strategy.experimental_distribute_dataset(train_dataset) if not args.random_data else None

        # Get model
        models = {
            "resnet":
            tf.keras.applications.ResNet50(
                include_top=True,
                weights="imagenet",
                classes=num_classes,
            ),
            "vgg":
            tf.keras.applications.VGG16(
                include_top=True,
                weights="imagenet",
                classes=num_classes,
            ),
            "alexnet":
            alexnet(input_shape, num_classes),
            "vit":
            vit(input_shape, num_classes),
        }
        model = models[args.model_name]
        print(model.summary())

        # Instantiate an optimizer.
        optimizer = keras.optimizers.SGD(learning_rate=1e-3, momentum=0.9)
        # Instantiate a loss function.
        loss_fn = keras.losses.SparseCategoricalCrossentropy(
            from_logits=True, reduction=tf.keras.losses.Reduction.NONE)

        # Prepare the metrics.
        train_acc_metric = keras.metrics.SparseCategoricalAccuracy()
        val_acc_metric = keras.metrics.SparseCategoricalAccuracy()

        train(distributed=distributed,
              epochs=2,
              model=model,
              train_dataset=train_dataset,
              optimizer=optimizer,
              loss_fn=loss_fn,
              train_acc_metric=train_acc_metric,
              val_acc_metric=val_acc_metric,
              strategy=strategy,
              random_data=args.random_data,
              batch_size=batch_size,
        )

    # Distributed training
    if distributed:
        with strategy.scope():
            run_training(
                batch_size=args.batch_size,
                distributed=distributed,
                strategy=strategy,
            )
    else:
        tf.config.threading.set_intra_op_parallelism_threads(10)
        tf.config.threading.set_inter_op_parallelism_threads(10)
        run_training(
            batch_size=args.batch_size,
            distributed=distributed,
        )

def get_hosts_list(filename, port=15000):
    with open(filename, 'r') as f:
        lines = f.readlines()
        lines = [f"{line.rstrip()}:{port}" for line in lines]
        return lines

if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        "PT Vision benchmarks - Proc-based distributed", parents=[]
    )
    parser.add_argument('--node_rank', default=0, type=int,
                        help='Node rank')
    parser.add_argument('--hosts_filepath', default=None, type=str,
                        help='file containing learnfair machines on which to run')
    parser.add_argument('--model_name', default='alexnet', type=str,
                        help='Model type [alexnet, vgg, resnet]')
    parser.add_argument('--batch_size', default=32, type=int,
                        help='Batch size')
    parser.add_argument('--random_data', action='store_true')
    args = parser.parse_args()

    if args.hosts_filepath:
        os.environ.pop('TF_CONFIG', None)
        hosts_list = get_hosts_list(args.hosts_filepath)
        print(f"Starting on hosts {hosts_list}")
        tf_config = {
            'cluster': {
                'worker': hosts_list
            },
            'task': {'type': 'worker', 'index': args.node_rank}
        }
        os.environ['TF_CONFIG'] = json.dumps(tf_config)
        nhosts = len(hosts_list)
        print(f"Will run training on {nhosts} machines and {nhosts * 8} gpu")
        args.distributed = True
    else:
        args.distributed = False
        
    main(args)
